#!/usr/bin/env python

"""
Simple built-in classes and functions.

Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013
              Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

Objects which provide code that shall always be compiled should provide
docstrings. Objects without code should be provided by native library code.

Classes without docstrings do not have instantiators generated for them.

Methods defined in classes are generated if they have docstrings, regardless of
whether their classes have docstrings.
"""

import native

class object:
    def __init__(self):
        "No-operation."
        pass
    def __bool__(self):
        "Objects are true by default."
        return True

class basestring(object):
    def __init__(self, x=None): pass
    def __contains__(self, value): pass

    def __getitem__(self, index):
        # Note usage.
        IndexError

    def __getslice__(self, start, end=None): pass

    def __iadd__(self, other):
        "Return a new string for the operation."
        return _binary_op(self, other, native._str_add)

    __add__ = __radd__ = __iadd__

    def __mul__(self, other): pass
    def __rmul__(self, other): pass
    def __mod__(self, other): pass
    def __rmod__(self, other): pass

    def __lt__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._str_lt)

    def __gt__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._str_gt)

    def __le__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__gt__(other))

    def __ge__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__lt__(other))

    def __eq__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._str_eq)

    def __ne__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__eq__(other))

    def __len__(self): pass
    def __str__(self): pass

    def __bool__(self):
        return _negate(native._str_eq(self, ""))

    def join(self, l): pass
    def split(self, s): pass
    def startswith(self, s): pass
    def endswith(self, s): pass

    def __iter__(self):

        "Return an iterator."

        return listiterator(self)

class bool(object):
    def __bool__(self):
        "Identity operation."
        return self
    def __str__(self): pass

class buffer(object):
    def __init__(self, size): pass
    def append(self, s): pass
    def __str__(self): pass

class complex(object):
    def __init__(self, real, imag=None): pass

class dict(object):
    def __init__(self, *args): pass
    def __setitem__(self, key, value): pass

    def __getitem__(self, key):
        # Note usage.
        KeyError

class file(object):
    def read(self, n=None): pass
    def write(self, s): pass
    def close(self): pass

class float(object):
    def __init__(self, number_or_string=None): pass
    def __iadd__(self, other): pass
    def __isub__(self, other): pass
    def __add__(self, other): pass
    def __radd__(self, other): pass
    def __sub__(self, other): pass
    def __rsub__(self, other): pass
    def __mul__(self, other): pass
    def __rmul__(self, other): pass
    def __div__(self, other): pass
    def __rdiv__(self, other): pass
    def __floordiv__(self, other): pass
    def __rfloordiv__(self, other): pass
    def __mod__(self, other): pass
    def __rmod__(self, other): pass
    def __pow__(self, other): pass
    def __rpow__(self, other): pass
    def __lt__(self, other): pass
    def __gt__(self, other): pass
    def __le__(self, other): pass
    def __ge__(self, other): pass
    def __eq__(self, other): pass
    def __ne__(self, other): pass
    def __neg__(self): pass
    def __pos__(self): pass
    def __str__(self): pass
    def __bool__(self): pass

class frozenset(object):
    def __init__(self, iterable): pass

class function(object):
    pass

class int(object):
    def __init__(self, number_or_string=None): pass

    def __iadd__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_add)

    def __isub__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_sub)

    def __imul__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_mul)

    def __idiv__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_div)

    def __imod__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_mod)

    def __ipow__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_pow)

    def __iand__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_and)

    def __ior__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_or)

    def __ixor__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_xor)

    __add__ = __radd__ = __iadd__
    __sub__ = __isub__

    def __rsub__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_rsub)

    __mul__ = __rmul__ = __imul__
    __div__ = __idiv__

    def __rdiv__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_rdiv)

    def __floordiv__(self, other): pass
    def __rfloordiv__(self, other): pass

    __mod__ = __imod__

    def __rmod__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_rmod)

    __pow__ = __ipow__

    def __rpow__(self, other):
        "Return a new int for the operation."
        return _binary_op(self, other, native._int_rpow)

    __and__ = __rand__ = __iand__
    __or__ = __ror__ = __ior__
    __xor__ = __rxor__ = __ixor__

    def __lt__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._int_lt)

    def __gt__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._int_gt)

    def __le__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__gt__(other))

    def __ge__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__lt__(other))

    def __eq__(self, other):
        "Return a new boolean for the comparison."
        return _binary_op(self, other, native._int_eq)

    def __ne__(self, other):
        "Return a new boolean for the comparison."
        return _negate(self.__eq__(other))

    def __invert__(self): pass
    def __neg__(self): pass
    def __pos__(self): pass
    def __str__(self): pass
    def __lshift__(self): pass
    def __rlshift__(self): pass
    def __rshift__(self): pass
    def __rrshift__(self): pass

    def __bool__(self):
        "Return whether this int is non-zero."
        return _negate(native._int_eq(self, 0))

class list(object):

    "Implementation of list."

    def __init__(self, args=None):

        "Initialise the list."

        self.__new__()

        if args is not None:
            self.extend(args)

    def __new__(self):
        # Reserve space for a fragment reference.
        self._elements = None

    def __getitem__(self, index):

        "Return the item or slice specified by 'index'."

        return _getitem(self, index)

    def __contains__(self, value): pass
    def __setitem__(self, index, value): pass

    def __getslice__(self, start, end=None):

        "Return a slice starting from 'start', with the optional 'end'."

        return _getslice(self, start, end)

    def __setslice__(self, start, end, slice): pass
    def append(self, value): pass

    def extend(self, iterable):

        "Extend the list with the contents of 'iterable'."

        for i in iterable:
            self.append(i)

    def pop(self): pass
    def reverse(self): pass
    def sort(self, cmp=None, key=None, reverse=0): pass
    def __len__(self): pass
    def __add__(self, other): pass
    def __iadd__(self, other): pass
    def __str__(self): pass

    def __bool__(self):

        "Lists are true if non-empty."

        return self.__len__() != 0

    def __iter__(self):

        "Return an iterator."

        return listiterator(self)

    # Special implementation methods.

    def __get_single_item__(self, index): pass

class listiterator(object):

    "Implementation of listiterator."

    def __init__(self, l):

        "Initialise with the given list 'l'."

        self.l = l
        self.i = 0

    def next(self):

        "Return the next item."

        try:
            value = self.l[self.i]
            self.i += 1
            return value
        except IndexError:
            raise StopIteration()

class long(object):
    def __init__(self, number_or_string=None): pass
    def __iadd__(self, other): pass
    def __isub__(self, other): pass
    def __add__(self, other): pass
    def __radd__(self, other): pass
    def __sub__(self, other): pass
    def __rsub__(self, other): pass
    def __mul__(self, other): pass
    def __rmul__(self, other): pass
    def __div__(self, other): pass
    def __rdiv__(self, other): pass
    def __floordiv__(self, other): pass
    def __rfloordiv__(self, other): pass
    def __and__(self, other): pass
    def __rand__(self, other): pass
    def __or__(self, other): pass
    def __ror__(self, other): pass
    def __xor__(self, other): pass
    def __rxor__(self, other): pass
    def __lt__(self, other): pass
    def __gt__(self, other): pass
    def __le__(self, other): pass
    def __ge__(self, other): pass
    def __eq__(self, other): pass
    def __ne__(self, other): pass
    def __neg__(self): pass
    def __pos__(self): pass
    def __str__(self): pass
    def __bool__(self): pass

class set(object):
    def __init__(self, iterable): pass
    def add(self, item): pass
    def clear(self): pass
    def copy(self): pass
    def difference(self, other): pass
    def difference_update(self, other): pass
    def discard(self, item): pass
    def intersection(self, other): pass
    def intersection_update(self, other): pass
    def issubset(self, other): pass
    def issuperset(self, other): pass
    def pop(self): pass
    def remove(self, item): pass
    def symmetric_difference(self, other): pass
    def symmetric_difference_update(self, other): pass
    def union(self, other): pass
    def update(self, other): pass

# See below for slice.

class str(basestring):
    pass

class type(object):
    pass

class tuple(object):

    #"Implementation of tuple."

    def __init__(self, args): pass

    def __getitem__(self, index):

        "Return the item or slice specified by 'index'."

        return _getitem(self, index)

    def __getslice__(self, start, end=None):

        "Return a slice starting from 'start', with the optional 'end'."

        return _tuple(_getslice(self, start, end))

    def __len__(self): pass
    def __add__(self, other): pass
    def __str__(self): pass

    def __bool__(self):

        "Tuples are true if non-empty."

        return self.__len__() != 0

    def __iter__(self):

        "Return an iterator."

        return listiterator(self)

    # Special implementation methods.

    def __get_single_item__(self, index): pass

class unicode(basestring):
    pass

class xrange(object):

    "Implementation of xrange."

    NO_END = object()

    def __init__(self, start_or_end, end=NO_END, step=1):

        "Initialise the xrange with the given 'start_or_end', 'end' and 'step'."

        if end is xrange.NO_END:
            self.start = 0
            self.end = start_or_end
        else:
            self.start = start_or_end
            self.end = end

        self.step = step
        self.current = self.start
        self.limited = self.end is not xrange.NO_END

    def __iter__(self):

        "Return an iterator, currently self."

        return self

    def next(self):

        "Return the next item or raise a StopIteration exception."

        if self.limited:
            if self.step < 0 and self.current <= self.end or self.step > 0 and self.current >= self.end:
                raise StopIteration()

        current = self.current
        self.current += self.step
        return current

class slice(xrange):

    "Implementation of slice."

    def __init__(self, start_or_end=None, end=xrange.NO_END, step=1):

        "Initialise the slice with the given 'start_or_end', 'end' and 'step'."

        xrange.__init__(self, start_or_end, end, step)

# Exceptions and warnings.

class BaseException(object):

    "Implementation of BaseException."

    def __init__(self, *args):
        self._pc = None # remember where the exception occurred
        self.args = args

class Exception(BaseException): pass
class Warning(object): pass

class ArithmeticError(Exception): pass
class AssertionError(Exception): pass
class AttributeError(Exception): pass
class DeprecationWarning(Exception): pass
class EOFError(Exception): pass
class EnvironmentError(Exception): pass
class FloatingPointError(Exception): pass
class FutureWarning(Warning): pass
class GeneratorExit(Exception): pass
class ImportError(Exception): pass
class ImportWarning(Warning): pass
class IndentationError(Exception): pass
class IndexError(Exception): pass
class IOError(Exception): pass
class KeyError(Exception): pass
class KeyboardInterrupt(Exception): pass
class LookupError(Exception): pass
class MemoryError(Exception): pass
class NameError(Exception): pass
class NotImplementedError(Exception): pass
class OSError(Exception): pass
class OverflowError(Exception): pass
class PendingDeprecationWarning(Warning): pass
class ReferenceError(Exception): pass
class RuntimeError(Exception): pass
class RuntimeWarning(Warning): pass
class StandardError(Exception): pass
class StopIteration(Exception): "Implementation of StopIteration."
class SyntaxError(Exception): pass
class SyntaxWarning(Warning): pass
class SystemError(Exception): pass
class SystemExit(Exception): pass
class TabError(Exception): pass
class TypeError(Exception): pass
class UnboundLocalError(Exception): pass
class UnicodeDecodeError(Exception): pass
class UnicodeEncodeError(Exception): pass
class UnicodeError(Exception): pass
class UnicodeTranslateError(Exception): pass
class UnicodeWarning(Warning): pass
class UserWarning(Warning): pass
class ValueError(Exception): pass
class ZeroDivisionError(Exception): pass

# Various types.

#class ellipsis: pass

class NoneType:
    def __bool__(self):
        "None is always false."
        return False

class NotImplementedType: pass

# General functions.
# NOTE: Some of these are actually provided by classes in CPython.
# NOTE: We may refuse to support some of these in practice, such as...
# NOTE: __import__, super, reload.

def abs(number): pass
def all(iterable): pass
def any(iterable): pass
def callable(obj): pass
def chr(i): pass
def classmethod(function): pass
def cmp(x, y): pass
def compile(source, filename, mode, flags=None, dont_inherit=None): pass
def delattr(obj, name): pass
def dir(obj=None): pass
def divmod(x, y): pass
def enumerate(iterable): pass
def eval(source, globals=None, locals=None): pass
def execfile(filename, globals=None, locals=None): pass
def filter(function, sequence): pass

_getattr_default=object() # a unique placeholder for a missing value
def getattr(obj, name, default=_getattr_default):

    "Implementation of getattr (activating the _accessor type)."

    _accessor # avoid isinstance but ensure that this class is included

    try:
        return _getattr(obj, name)
    except AttributeError:
        if default is not _getattr_default:
            return default
        else:
            raise

def globals(): pass
def hasattr(obj, name): pass
def hash(obj): pass
def help(*args): pass
def hex(number): pass
def id(obj): pass
def input(prompt=None): pass

def isinstance(obj, cls_or_tuple):

    """
    Return whether 'obj' is an instance of 'cls_or_tuple', where the latter is
    either a class or a tuple of classes.
    """

    # NOTE: CPython insists on tuples, but any sequence might be considered
    # NOTE: acceptable.

    if _isinstance(cls_or_tuple, tuple):
        for cls in cls_or_tuple:
            if obj.__class__ is cls or _isinstance(obj, cls):
                return True
        return False
    else:
        return obj.__class__ is cls_or_tuple or _isinstance(obj, cls_or_tuple)

def issubclass(obj, cls_or_tuple): pass

def iter(collection):

    "Implementation of iter without callable plus sentinel support."

    return collection.__iter__()

def len(obj):

    "Implementation of len."

    return obj.__len__()

def locals(): pass

def map(function, *args): pass

def max(*args):

    "Implementation of max."

    highest = args[0]
    for arg in args[1:]:
        if arg > highest:
            highest = arg
    return highest

def min(*args):

    "Implementation of min."

    lowest = args[0]
    for arg in args[1:]:
        if arg > lowest:
            lowest = arg
    return lowest

def oct(number): pass
def open(name, mode=None, buffering=None): pass
def ord(c): pass
def pow(x, y, z=None): pass
def property(fget=None, fset=None, fdel=None, doc=None): pass

def range(start_or_end, end=None, step=1):

    "Implementation of range."

    return list(xrange(start_or_end, end, step))

def raw_input(prompt=None): pass
def reduce(function, sequence, initial=None): pass
def reload(module): pass
def repr(obj): pass
def reversed(sequence): pass
def round(number, ndigits=None): pass
def setattr(obj, name, value): pass
def sorted(iterable, cmp=None, key=None, reverse=False): pass
def staticmethod(function): pass
def sum(sequence, start=0): pass
def super(*args): pass
def unichr(i): pass
def vars(obj=None): pass
def zip(*args): pass

# Utility functions.

def _binary_op(self, other, op):

    "Test the type of 'other' and perform 'op'."

    if self.__class__ is other.__class__:
        return op(self, other)
    else:
        return NotImplemented

def _negate(result):

    "Negate any valid logical value."

    if result is NotImplemented:
        return result
    else:
        return not result

def _get_absolute_index(index, length):

    """
    Return the absolute index for 'index' given a collection having the
    specified 'length'.
    """

    if index < 0:
        return length + index
    else:
        return index

def _normalise_index(index, length):

    "Normalise 'index' for a collection having the specified 'length'."

    return _min(length, _max(0, _get_absolute_index(index, length)))

def _max(x, y):

    "Return the maximum of 'x' and 'y'."

    if x >= y:
        return x
    else:
        return y

def _min(x, y):

    "Return the minimum of 'x' and 'y'."

    if x <= y:
        return x
    else:
        return y

def _getitem(seq, index):

    "Return the item or slice specified by 'index'."

    if isinstance(index, int):
        index = _normalise_index(index, len(seq))
        return seq.__get_single_item__(index)
    elif isinstance(index, slice):
        return seq.__getslice__(index.start, index.end)
    else:
        raise TypeError

def _getslice(seq, start, end=None):

    "Return a slice starting from 'start', with the optional 'end'."

    length = len(seq)

    if start is None:
        start = 0
    else:
        start = _normalise_index(start, length)

    if end is None:
        end = length
    else:
        end = _normalise_index(end, length)

    result = []
    while start < end:
        result.append(seq.__get_single_item__(start))
        start += 1
    return result

# Special implementation classes.

class _accessor(str):

    "A string which can be used to access attributes."

    def __new__(self):
        # Reserve space for an object table index.
        self._index = None

# Special implementation functions.

def _getattr(obj, name): pass
def _isinstance(obj, cls): pass
def _print(dest, *args): pass
def _printnl(dest, *args): pass
def _tuple(l): pass

# Reference some names to ensure their existence. This should be everything
# mentioned in a get_builtin or load_builtin call. Instances from this module
# should be predefined constants.

function
AttributeError
#IndexError
NoneType
NotImplementedType
#StopIteration
TypeError

bool
#ellipsis
#list
tuple
type
#xrange

# vim: tabstop=4 expandtab shiftwidth=4
